/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
#ifndef UTILITY_MONITORS_H_
#define UTILITY_MONITORS_H_

#include "assert.h"
#include <cmath>
#include <functional>
#include "hash.h"
#include <string>
#include <sstream>

class H3HashFamily;
typedef uint64_t Address;

class LRUMon {
    private:
        uint32_t umonLines;
        double samplingFactor;
        Address MAX_LIMIT = 1ull<<32;
        inline Address mask(Address hashed) { return hashed & (MAX_LIMIT - 1); }
        Address limit;         // Addresses whose hash lie above this will be ignored
        uint32_t umonBuckets; //umon ways
        uint32_t sets; //umon sets. Should be power of 2.
        uint64_t setsBits;

        uint64_t* curWayHits;
        uint64_t curMisses;

        //Even for high associativity/number of buckets, performance of this is not important because we downsample so much (so this is a LL)
        struct Node {
            Address addr;
            struct Node* next;
        };
        Node** array;
        Node** heads;

        H3HashFamily* hf;

    public:
        LRUMon(uint32_t _bankLines, uint32_t _umonLines, uint32_t _umonBuckets) {
            umonLines = _umonLines;
            umonBuckets = _umonBuckets;
            samplingFactor = 1.*_bankLines/umonLines;
            printf("UMON: Sampling factor: %f = %u / %u\n", samplingFactor, _bankLines, umonLines);
            sets = umonLines/umonBuckets;

            heads = (Node**)calloc(sets, sizeof(Node*));
            array = (Node**)calloc(sets, sizeof(Node*));
            for (uint32_t i = 0; i < sets; i++) {
                array[i] = (Node*)calloc(umonBuckets, sizeof(Node));
                heads[i] = &array[i][0];
                for (uint32_t j = 0; j < umonBuckets-1; j++) {
                    array[i][j].next = &array[i][j+1];
                }
            }

            curWayHits = (uint64_t*) calloc(umonBuckets, sizeof(uint64_t));
            curMisses = 0;

            hf = new H3HashFamily(2, 32, 0xF000BAAD);

            setScale(1.);

            setsBits = 0;
            auto tmp = sets;
            while (tmp >>= 1) setsBits++;
        }

        ~LRUMon () {};

        void setScale(double scale) {
            assert(scale > 0);
            limit = MAX_LIMIT * (scale / samplingFactor);
            limit = std::min(limit, MAX_LIMIT);
        }

        double getScale() const {
            return double(limit) * samplingFactor / double(MAX_LIMIT);
        }

        void access(Address lineAddr) {
            //1. Hash to decide if it should go in the cache
            Address hashed = hf->hash(0, lineAddr);
            if (mask(hashed) > limit) return;

            //2. Insert; hit or miss?
            uint64_t setMask = ~(((uint64_t)-1LL) << setsBits);
            uint64_t set = (hf->hash(1, lineAddr)) & setMask;

            // Check hit
            Node* prev = nullptr;
            Node* cur = heads[set];
            bool hit = false;
            for (uint32_t b = 0; b < umonBuckets; b++) {
                if (cur->addr == lineAddr) { //Hit at position b, profile
                    curWayHits[b]++;
                    hit = true;
                    break;
                } else if (b < umonBuckets-1) {
                    prev = cur;
                    cur = cur->next;
                }
            }

            //Profile miss, kick cur out, put lineAddr in
            if (!hit) {
                curMisses++;
                assert(cur->next == nullptr);
                cur->addr = lineAddr;
            }

            //Move cur to MRU (happens regardless of whether this is a hit or a miss)
            if (prev) {
                prev->next = cur->next;
                cur->next = heads[set];
                heads[set] = cur;
            }
        }

        uint64_t getNumAccesses() const {
            uint64_t total = curMisses;
            for (uint32_t i = 0; i < umonBuckets; i++) {
                total += curWayHits[umonBuckets - i - 1];
            }
            return total * samplingFactor;
        }

        void getMisses(uint64_t* misses) {
            uint64_t total = curMisses;
            for (uint32_t i = 0; i < umonBuckets; i++) {
                misses[umonBuckets - i] = total * samplingFactor;
                total += curWayHits[umonBuckets - i - 1];
            }
            misses[0] = total * samplingFactor;
        }

        void startNextInterval() {
            curMisses = 0;
            for (uint32_t b = 0; b < umonBuckets; b++) {
                curWayHits[b] = 0;
            }
        }

        uint32_t getUmonBuckets() const { return umonBuckets; }
};

#endif  // UTILITY_MONITORS_H_
