'''
/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
'''

### Routines to manipulate miss curves ###
import numpy as np

def convexifyCurve(curve):
    missCurveLength = len(curve)
    xvals = np.arange(missCurveLength)
    hull = np.empty(missCurveLength)
    hull[0] = 0
    hullSize = 1

    for i in range(1, missCurveLength):
        x2 = xvals[i]
        y2 = curve[i]

        while hullSize > 1:
            i1 = hull[hullSize - 1]
            i0 = hull[hullSize - 2]

            x1 = xvals[i1]
            y1 = curve[i1]
            x0 = xvals[i0]
            y0 = curve[i0]

            remove = abs( (y2 - y1) * (x2 - x0) ) >= abs( (y2 - y0) * (x2 - x1) )
            if remove:
                hullSize -= 1
            else:
                break

        hull[hullSize] = i
        hullSize += 1

    hull = hull[:hullSize]
    convexCurve = np.interp(xvals, xvals[hull.tolist()], curve[hull.tolist()])
    return convexCurve

# Assumes convex curves, uses hill climbing in one single pass
def partition(curve1, curve2):
    n = len(curve1)
    allocs = np.zeros((n,2), dtype=int)

    for s in range(1, n):
        allocs[s] = allocs[s-1]
        slope1 = curve1[allocs[s,0]] - curve1[allocs[s,0] + 1]
        slope2 = curve2[allocs[s,1]] - curve2[allocs[s,1] + 1]
        if slope1 > slope2:
            bestPart = 0
        else:
            bestPart = 1

        allocs[s,bestPart] += 1

    return allocs

def systemMissCurve(curve1, curve2):
    allocs = partition(curve1, curve2)
    newCurve = curve1[ allocs[:,0] ] + curve2[ allocs[:,1] ]
    return newCurve

def combinedMissCurve(curve1, curve2):
    missCurveLength = len(curve1)
    xvals = np.arange(missCurveLength)

    index1 = 0.0
    index2 = 0.0

    newCurve = np.zeros(missCurveLength)
    newCurve[0] = curve1[0] + curve2[0]

    for i in range(1, missCurveLength):
        v1 = np.interp(index1, xvals, curve1)
        v2 = np.interp(index2, xvals, curve2)

        if v1 + v2 == 0.0:
            break

        # ratio of accesses
        index1 += v1/(v1+v2)
        index2 += v2/(v1+v2)

        m1 = np.interp([index1], xvals, curve1)
        m2 = np.interp([index2], xvals, curve2)

        newCurve[i] = m1 + m2

    return newCurve

### Routines that do not assume convex curves ###
def lookahead(balance, curve1, curve2):
    allocs = np.zeros(2, dtype=int)
    assert(len(curve1) == len(curve2))
    missCurvePoints = len(curve1)

    missCurves = np.zeros((2, missCurvePoints), dtype=int)
    missCurves[0,:] = curve1
    missCurves[1,:] = curve2

    while balance > 0:
        bestSlope = 0
        bestPart = -1
        bestAlloc = 0
        for p in [0, 1]:
            curve = missCurves[p]
            for s in range(allocs[p] + 1, missCurvePoints):
                if s - allocs[p] > balance:
                    break;
                benefit = float(curve[allocs[p]] - curve[s])
                size = s - allocs[p]
                slope = benefit / size
                if slope >= bestSlope :
                    bestSlope = slope
                    bestPart = p
                    bestAlloc = s - allocs[p]

        if bestPart == -1 :
            break
        else :
            allocs[bestPart] += bestAlloc;
            balance -= bestAlloc;

    m1 = curve1[allocs[0]]
    m2 = curve2[allocs[1]]

    return m1 + m2

'''
def systemMissCurve(curve1, curve2):
    missCurveLength = len(curve1)

    newCurve = np.zeros(missCurveLength)
    newCurve[0] = curve1[0] + curve2[0]

    for balance in range(1, missCurveLength):
        newCurve[balance] = lookahead(balance, curve1, curve2)

    return newCurve
'''
